@@ -12,6 +12,7 @@
 #ifdef ICA_FIPS
 
 #include <errno.h>
+#include <fcntl.h>
 #include <openssl/crypto.h>
 #include <openssl/evp.h>
 #include <stdint.h>
@@ -19,8 +20,11 @@
 #include <stdlib.h>
 #include <string.h>
 #include <syslog.h>
+#include <unistd.h>
 #include <dlfcn.h>
 #include <link.h>
+#include <sys/mman.h>
+#include <sys/stat.h>
 
 #include <openssl/opensslconf.h>
 #ifdef OPENSSL_FIPS
@@ -76,6 +80,7 @@ int fips;
 
 #define LIBICA_FIPS_CONFIG		LIBICA_CONFDIR "/libica/openssl3-fips.cnf"
 
+static int drbg_kat(void);
 static int aes_ecb_kat(void);
 static int aes_cbc_kat(void);
 static int aes_cbc_cs_kat(void);
@@ -87,15 +92,44 @@ static int aes_gcm_kat(void);
 static int aes_xts_kat(void);
 static int aes_cmac_kat(void);
 
-static int des3_ecb_kat(void);
-static int des3_cbc_kat(void);
-static int des3_cbc_cs_kat(void);
-static int des3_cfb_kat(void);
-static int des3_ofb_kat(void);
-static int des3_ctr_kat(void);
-static int des3_cmac_kat(void);
-
 static int rsa_kat(void);
+static int ecdsa_kat(void);
+static int ecdh_kat(void);
+
+static int function_supported_via_cpacf(unsigned int func)
+{
+	ica_adapter_handle_t ah;
+	libica_func_list_element* libica_func_list = NULL;
+	unsigned int count;
+	size_t i;
+	int ret = 0;
+
+	if (ica_open_adapter(&ah))
+		return ret;
+
+	if (ica_get_functionlist(NULL, &count) != 0)
+		return ret;
+
+	libica_func_list = malloc(sizeof(libica_func_list_element) * count);
+	if (!libica_func_list)
+		return ret;
+
+	if (ica_get_functionlist(libica_func_list, &count) != 0)
+		goto done;
+
+	for (i = 0; i < count; i++) {
+		if (libica_func_list[i].mech_mode_id == func &&
+			libica_func_list[i].flags == ICA_FLAG_SHW) {
+			ret = 1;
+			goto done;
+		}
+	}
+
+done:
+	free(libica_func_list);
+	ica_close_adapter(ah);
+	return ret;
+}
 
 #ifndef NO_CPACF
 #define SHA_KAT(_sha_, _ctx_)						\
@@ -103,6 +137,8 @@ static int sha##_sha_##_kat(void) {					\
 	sha##_ctx_##_context_t ctx;					\
 	size_t i;							\
 	unsigned char out[SHA##_sha_##_HASH_LENGTH];			\
+	if (!function_supported_via_cpacf(SHA1))			\
+		return 0;										\
 	for (i = 0; i < SHA##_sha_##_TV_LEN; i++) {			\
 		if (ica_sha##_sha_(SHA_MSG_PART_ONLY,			\
 		    SHA##_sha_##_TV[i].msg_len, SHA##_sha_##_TV[i].msg,	\
@@ -127,6 +163,8 @@ static int sha##_sha_##_kat(void) {					\
 	sha##_ctx_##_context_t ctx;					\
 	size_t i;							\
 	unsigned char out[SHA##_sha_##_HASH_LENGTH];			\
+	if (!function_supported_via_cpacf(SHA1))			\
+		return 0;										\
 	for (i = 0; i < SHA##_sha_##_TV_LEN; i++) {			\
 		if (ica_sha##_sha_(SHA_MSG_PART_ONLY,			\
 		    SHA##_sha_##_TV[i].msg_len, SHA##_sha_##_TV[i].msg,	\
@@ -155,7 +193,7 @@ static inline int sha3_available(void)
 	rc = ica_sha3_224(SHA_MSG_PART_ONLY, sizeof(test_data), test_data,
 			&sha3_224_context, output_hash);
 
-	return (rc == ENODEV ? 0 : 1);
+	return (rc != 0 ? 0 : 1);
 }
 
 #ifndef NO_CPACF
@@ -213,6 +251,7 @@ fips_init(void)
 {
 	FILE *fd;
 	char fips_flag;
+	char *fips_override;
 
 	if ((fd = fopen(FIPS_FLAG, "r")) == NULL)
 		return;
@@ -223,6 +262,12 @@ fips_init(void)
 	}
 	fclose(fd);
 
+	/* Allow to override the kernel fips indication for testing on
+	 * non-fips systems. */
+	fips_override = getenv("LIBICA_FIPS_FLAG");
+	if ((fips_override != NULL) && (atoi(fips_override) == 1))
+		fips_flag = '1';
+
 	if (fips_flag - '0') {
 #if !OPENSSL_VERSION_PREREQ(3, 0)
 		/* Set libica into FIPS mode. */
@@ -235,12 +280,18 @@ fips_init(void)
 		FIPS_mode_set(1);
 #else
 		fips = 0;
+
+#ifndef NO_FIPS_CONFIG_LOAD
+		/* Allow to skip reading the openssl 3.x fips config. Tests showed
+		 * that this step must be skipped on RHEL9 systems. But on other
+		 * systems, or with a locally built openssl, this step is necessary. */
 		if (!OSSL_LIB_CTX_load_config(openssl_libctx, LIBICA_FIPS_CONFIG)) {
 			syslog(LOG_ERR, "Libica failed to load openssl fips config %s\n",
 					LIBICA_FIPS_CONFIG);
 			fips |= ICA_FIPS_INTEGRITY;
 			return;
 		}
+#endif
 
 		openssl_provider = OSSL_PROVIDER_load(openssl_libctx, "fips");
 		if (openssl_provider == NULL) {
@@ -297,22 +348,57 @@ static char *make_hmac_path(const char *origpath)
 	return path;
 }
 
-static int compute_file_hmac(const char *path, void **buf, size_t *hmaclen)
+static EVP_PKEY *get_pkey(void)
 {
-	FILE *fp = NULL;
-	int rc = -1;
-	unsigned char rbuf[READ_BUFFER_LENGTH];
 	unsigned char *keybuf;
-	EVP_MD_CTX *mdctx = NULL;
-	EVP_PKEY *pkey = NULL;
-	size_t hlen, len;
 	long keylen;
-
-	*buf = NULL;
-	*hmaclen = 0;
+	EVP_PKEY *pkey = NULL;
 
 	keybuf = OPENSSL_hexstr2buf(hmackey, &keylen);
+	if (keybuf == NULL)
+		goto end;
+
 	pkey = EVP_PKEY_new_mac_key(EVP_PKEY_HMAC, NULL, keybuf, (int)keylen);
+end:
+	if (keybuf) {
+		OPENSSL_cleanse(keybuf, keylen);
+		OPENSSL_free(keybuf);
+	}
+	return pkey;
+}
+
+static void* mmap_file(const char *path, struct stat *statbuf)
+{
+	int fd;
+	void *ptr = NULL;
+
+	fd = open(path, O_RDONLY);
+	if (fd < 0)
+		goto end;
+
+	if (fstat(fd, statbuf) < 0)
+		goto end_close;
+
+	ptr = mmap(0, statbuf->st_size, PROT_READ, MAP_SHARED, fd, 0);
+	if (ptr == MAP_FAILED)
+		ptr = NULL;
+end_close:
+	close(fd);
+end:
+	return ptr;
+}
+
+static int compute_file_hmac(const char *path, void **buf, size_t *hmaclen)
+{
+	int rc = -1;
+	unsigned char tmp[32];
+	size_t tmp_len = sizeof(tmp);
+	EVP_MD_CTX *mdctx = NULL;
+	EVP_PKEY *pkey = NULL;
+	void *fdata = NULL;
+	struct stat fdata_stat;
+
+	pkey = get_pkey();
 	if (!pkey)
 		goto end;
 
@@ -320,42 +406,36 @@ static int compute_file_hmac(const char *path, void **buf, size_t *hmaclen)
 	if (!mdctx)
 		goto end;
 
-	fp = fopen(path, "r");
-	if (fp == NULL)
+	fdata = mmap_file(path, &fdata_stat);
+	if (fdata == NULL)
 		goto end;
 
-	if (EVP_DigestSignInit(mdctx, NULL, EVP_sha256(), NULL, pkey) <= 0)
+	if (EVP_DigestSignInit(mdctx, NULL, EVP_sha256(), NULL, pkey) != 1)
 		goto end;
 
-	while ((len = fread(rbuf, 1, sizeof(rbuf), fp)) != 0) {
-		if (EVP_DigestSignUpdate(mdctx, rbuf, len) <= 0) {
-			goto end;
-		}
-	}
-
-	hlen = sizeof(rbuf);
-	if (EVP_DigestSignFinal(mdctx, rbuf, &hlen) <= 0)
+	if (EVP_DigestSign(mdctx, tmp, &tmp_len, fdata, fdata_stat.st_size) != 1)
 		goto end;
 
-	*buf = malloc(hlen);
+	*buf = malloc(tmp_len);
 	if (*buf == NULL)
 		goto end;
 
-	*hmaclen = hlen;
-
-	memcpy(*buf, rbuf, hlen);
+	*hmaclen = tmp_len;
+	memcpy(*buf, tmp, tmp_len);
 
 	rc = 0;
 
 end:
+	if (fdata != NULL)
+		munmap(fdata, fdata_stat.st_size);
 
 	if (pkey != NULL)
 		EVP_PKEY_free(pkey);
 
-	OPENSSL_free(keybuf);
-	EVP_MD_CTX_destroy(mdctx);
-	if (fp)
-		fclose(fp);
+	if (mdctx != NULL)
+		EVP_MD_CTX_destroy(mdctx);
+
+	OPENSSL_cleanse(tmp, sizeof(tmp));
 
 	return rc;
 }
@@ -385,6 +465,8 @@ static int load_known_hmac(const char *path, unsigned char **hmac, long *hmaclen
 	rc = 0;
 end:
 	fclose(fp);
+	if (known_hmac_str)
+		OPENSSL_cleanse(known_hmac_str, strlen(known_hmac_str));
 	free(known_hmac_str);
 	free(hmacpath);
 
@@ -401,9 +483,9 @@ static int FIPSCHECK_verify(const char *path)
 {
 	int rc = 0;
 	unsigned char *known_hmac = NULL;
-	long known_hmac_len;
+	long known_hmac_len = 0;
 	void *computed_hmac = NULL;
-	size_t computed_hmac_len;
+	size_t computed_hmac_len = 0;
 
 	if (load_known_hmac(path, &known_hmac,  &known_hmac_len) != 0)
 		goto end;
@@ -419,6 +501,10 @@ static int FIPSCHECK_verify(const char *path)
 
 	rc = 1;
 end:
+	if (computed_hmac)
+		OPENSSL_cleanse(computed_hmac, computed_hmac_len);
+	if (known_hmac)
+		OPENSSL_cleanse(known_hmac, known_hmac_len);
 	free(computed_hmac);
 	OPENSSL_free(known_hmac);
 
@@ -500,30 +586,21 @@ static void fips_lib_integrity_check(void)
 void
 fips_powerup_tests(void)
 {
-#ifdef NO_CPACF
-	/* 27 out of the 28 tests return EPERM if CPACF is disabled via config.
-	 * The rsa_kat() is not affected. */
-	int num_cpacf_tests = 27;
-#endif
-	int rc;
+	typedef int (*kat_func)(void);
+	kat_func kats[] = {
+		drbg_kat, sha1_kat, sha224_kat, sha256_kat, sha384_kat, sha512_kat,
+		sha3_224_kat, sha3_256_kat, sha3_384_kat, sha3_512_kat, aes_ecb_kat,
+		aes_cbc_kat, aes_cbc_cs_kat, aes_cfb_kat, aes_ctr_kat, aes_ofb_kat,
+		aes_ccm_kat, aes_gcm_kat, aes_xts_kat, aes_cmac_kat, rsa_kat,
+		ecdsa_kat, ecdh_kat,
+	};
+	size_t i, num_kats = sizeof(kats) / sizeof(kat_func);
 
-	/* Cryptographic algorithm test. */
-	rc = ica_drbg_health_test(ica_drbg_generate, 256, true, ICA_DRBG_SHA512)
-	    + sha1_kat() + sha224_kat() + sha256_kat() + sha384_kat()
-	    + sha512_kat() + sha3_224_kat() + sha3_256_kat() + sha3_384_kat()
-	    + sha3_512_kat() + des3_ecb_kat() + des3_cbc_kat()
-	    + des3_cbc_cs_kat() + des3_cfb_kat() + des3_ofb_kat()
-	    + des3_ctr_kat() + des3_cmac_kat() + aes_ecb_kat()
-	    + aes_cbc_kat() + aes_cbc_cs_kat() + aes_cfb_kat()
-	    + aes_ctr_kat() + aes_ofb_kat() + aes_ccm_kat() + aes_gcm_kat()
-	    + aes_xts_kat() + aes_cmac_kat() + rsa_kat();
-#ifndef NO_CPACF
-	if (rc != 0) {
-#else
-	if (rc != 0 && rc != num_cpacf_tests * EPERM) {
-#endif
-		fips |= ICA_FIPS_CRYPTOALG;
-		return;
+	for (i = 0; i < num_kats; i++) {
+		if (kats[i]() != 0) {
+			fips |= ICA_FIPS_CRYPTOALG;
+			return;
+		}
 	}
 
 /* ICA internal test does not link against the library. So we should
@@ -535,12 +612,23 @@ fips_powerup_tests(void)
 #endif
 }
 
+static int drbg_kat(void)
+{
+	if (!function_supported_via_cpacf(SHA512_DRNG))
+		return 0;
+
+	return ica_drbg_health_test(ica_drbg_generate, 256, true, ICA_DRBG_SHA512);
+}
+
 static int
 aes_ecb_kat(void) {
 	const struct aes_ecb_tv *tv;
 	size_t i;
 	unsigned char *out;
 
+	if (!function_supported_via_cpacf(AES_ECB))
+		return 0;
+
 	for (i = 0; i < AES_ECB_TV_LEN; i++) {
 		tv = &AES_ECB_TV[i];
 
@@ -574,6 +662,9 @@ aes_cbc_kat(void) {
 	size_t i;
 	unsigned char iv[AES_BLKSIZE], *out;
 
+	if (!function_supported_via_cpacf(AES_CBC))
+		return 0;
+
 	for (i = 0; i < AES_CBC_TV_LEN; i++) {
 		tv = &AES_CBC_TV[i];
 
@@ -613,6 +704,9 @@ aes_cbc_cs_kat(void)
 	size_t i;
 	unsigned char iv[AES_BLKSIZE], *out;
 
+	if (!function_supported_via_cpacf(AES_CBC_CS))
+		return 0;
+
 	for (i = 0; i < AES_CBC_CS_TV_LEN; i++) {
 		tv = &AES_CBC_CS_TV[i];
 
@@ -652,6 +746,9 @@ aes_cfb_kat(void) {
 	size_t i;
 	unsigned char iv[AES_BLKSIZE], *out;
 
+	if (!function_supported_via_cpacf(AES_CFB))
+		return 0;
+
 	for (i = 0; i < AES_CFB_TV_LEN; i++) {
 		tv = &AES_CFB_TV[i];
 
@@ -690,6 +787,9 @@ aes_ofb_kat(void) {
 	size_t i;
 	unsigned char iv[AES_BLKSIZE], *out;
 
+	if (!function_supported_via_cpacf(AES_OFB))
+		return 0;
+
 	for (i = 0; i < AES_OFB_TV_LEN; i++) {
 		tv = &AES_OFB_TV[i];
 
@@ -728,6 +828,9 @@ aes_ctr_kat(void) {
 	size_t i;
 	unsigned char *out, ctr[AES_BLKSIZE];
 
+	if (!function_supported_via_cpacf(AES_CTR))
+		return 0;
+
 	for (i = 0; i < AES_CTR_TV_LEN; i++) {
 		tv = &AES_CTR_TV[i];
 
@@ -766,6 +869,9 @@ aes_ccm_kat(void) {
 	size_t i;
 	unsigned char *ciphertext, *payload;
 
+	if (!function_supported_via_cpacf(AES_CCM))
+		return 0;
+
 	for (i = 0; i < AES_CCM_TV_LEN; i++) {
 		tv = &AES_CCM_TV[i];
 
@@ -810,6 +916,9 @@ aes_gcm_kat(void) {
 	unsigned char *out, *tag, icb[AES_BLKSIZE], ucb[AES_BLKSIZE],
 	    subkey[AES_BLKSIZE];
 
+	if (!function_supported_via_cpacf(AES_GCM))
+		return 0;
+
 	for (i = 0; i < AES_GCM_TV_LEN; i++) {
 		tv = &AES_GCM_TV[i];
 
@@ -819,14 +928,14 @@ aes_gcm_kat(void) {
 		if (tag == NULL || out == NULL)
 			goto _err_;
 
-		if ((ica_aes_gcm(out, tv->len, tv->ciphertext, tv->iv,
+		if ((ica_aes_gcm_internal(out, tv->len, tv->ciphertext, tv->iv,
 		    tv->ivlen, tv->aad, tv->aadlen, tv->tag, tv->taglen,
 		    tv->key, tv->keylen, ICA_DECRYPT) != tv->rv)
 		    || ((tv->rv == 0)
 		    && memcmp(tv->plaintext, out, tv->len)))
 			goto _err_;
 
-		if ((tv->rv == 0) && (ica_aes_gcm(tv->plaintext, tv->len,
+		if ((tv->rv == 0) && (ica_aes_gcm_internal(tv->plaintext, tv->len,
 		    out, tv->iv, tv->ivlen, tv->aad, tv->aadlen, tag,
 		    tv->taglen, tv->key, tv->keylen, ICA_ENCRYPT)
 		    || memcmp(tv->tag, tag, tv->taglen)
@@ -858,7 +967,7 @@ aes_gcm_kat(void) {
 			goto _err_;
 
 		memset(tag, 0, AES_BLKSIZE);
-		if (ica_aes_gcm_initialize(tv->iv, tv->ivlen, tv->key,
+		if (ica_aes_gcm_initialize_internal(tv->iv, tv->ivlen, tv->key,
 		    tv->keylen, icb, ucb, subkey, ICA_DECRYPT)
 		    || ica_aes_gcm_intermediate(out, tv->len - lastlen,
 		    tv->ciphertext, ucb, tv->aad, tv->aadlen, tag,
@@ -877,7 +986,7 @@ aes_gcm_kat(void) {
 		memset(icb, 0, sizeof(icb));
 		memset(icb, 0, sizeof(ucb));
 		memset(subkey, 0, sizeof(subkey));
-		if ((tv->rv == 0) && (ica_aes_gcm_initialize(tv->iv, tv->ivlen,
+		if ((tv->rv == 0) && (ica_aes_gcm_initialize_internal(tv->iv, tv->ivlen,
 		    tv->key, tv->keylen, icb, ucb, subkey, ICA_ENCRYPT)
 		    || ica_aes_gcm_intermediate(tv->plaintext,
 		    tv->len - lastlen, out, ucb, tv->aad, tv->aadlen, tag,
@@ -911,6 +1020,9 @@ aes_xts_kat(void) {
 	size_t i;
 	unsigned char *out, tweak[16];
 
+	if (!function_supported_via_cpacf(AES_XTS))
+		return 0;
+
 	for (i = 0; i < AES_XTS_TV_LEN; i++) {
 		tv = &AES_XTS_TV[i];
 
@@ -950,6 +1062,9 @@ aes_cmac_kat(void)
 	size_t i, lastlen;
 	unsigned char *mac, iv[AES_BLKSIZE];
 
+	if (!function_supported_via_cpacf(AES_CMAC))
+		return 0;
+
 	for (i = 0; i < AES_CMAC_TV_LEN; i++) {
 		tv = &AES_CMAC_TV[i];
 
@@ -1013,296 +1128,123 @@ _err_:
 }
 
 static int
-des3_ecb_kat(void) {
-	const struct des3_ecb_tv *tv;
-	size_t i;
-	unsigned char *out;
-
-	for (i = 0; i < DES3_ECB_TV_LEN; i++) {
-		tv = &DES3_ECB_TV[i];
-
-		out = malloc(tv->len);
-
-		if (out == NULL)
-			goto _err_;
-
-		if (ica_3des_ecb(tv->plaintext, out, tv->len,
-		    (unsigned char *)tv->key, ICA_ENCRYPT)
-		    || memcmp(tv->ciphertext, out, tv->len)
-		    || ica_3des_ecb(tv->ciphertext, out, tv->len,
-		    (unsigned char *)tv->key, ICA_DECRYPT)
-		    || memcmp(tv->plaintext, out, tv->len))
-			goto _err_;
-
-		free(out);
-	}
-	return 0;
-
-_err_:
-	free(out);
-#ifndef NO_CPACF
-	syslog(LOG_ERR, "Libica 3DES-ECB test failed.");
-#endif
-	return 1;
-}
-
-static int
-des3_cbc_kat(void) {
-	const struct des3_cbc_tv *tv;
-	size_t i;
-	unsigned char iv[DES3_BLKSIZE], *out;
-
-	for (i = 0; i < DES3_CBC_TV_LEN; i++) {
-		tv = &DES3_CBC_TV[i];
-
-		out = malloc(tv->len);
-
-		if (out == NULL)
-			goto _err_;
-
-		memcpy(iv, tv->iv, DES3_BLKSIZE);
-		if (ica_3des_cbc(tv->plaintext, out, tv->len,
-		    (unsigned char *)tv->key, iv, ICA_ENCRYPT)
-		    || memcmp(tv->ciphertext, out, tv->len))
-			goto _err_;
-
-		memcpy(iv, tv->iv, DES3_BLKSIZE);
-		if (ica_3des_cbc(tv->ciphertext, out, tv->len,
-		    (unsigned char *)tv->key, iv, ICA_DECRYPT)
-		    || memcmp(tv->plaintext, out, tv->len))
-			goto _err_;
-
-		free(out);
-	}
-	return 0;
-
-_err_:
-	free(out);
-#ifndef NO_CPACF
-	syslog(LOG_ERR, "Libica 3DES-CBC test failed.");
-#endif
-	return 1;
-}
-
-static int
-des3_cbc_cs_kat(void) {
-	const struct des3_cbc_cs_tv *tv;
-	size_t i;
-	unsigned char iv[DES3_BLKSIZE], *out;
-
-	for (i = 0; i < DES3_CBC_CS_TV_LEN; i++) {
-		tv = &DES3_CBC_CS_TV[i];
-
-		out = malloc(tv->len);
-
-		if (out == NULL)
-			goto _err_;
-
-		memcpy(iv, tv->iv, DES3_BLKSIZE);
-		if (ica_3des_cbc_cs(tv->plaintext, out, tv->len,
-		    (unsigned char *)tv->key, iv, ICA_ENCRYPT, tv->variant)
-		    || memcmp(tv->ciphertext, out, tv->len))
-			goto _err_;
-
-		memcpy(iv, tv->iv, DES3_BLKSIZE);
-		if (ica_3des_cbc_cs(tv->ciphertext, out, tv->len,
-		    (unsigned char *)tv->key, iv,
-		    ICA_DECRYPT, tv->variant)
-		    || memcmp(tv->plaintext, out, tv->len))
-			goto _err_;
-
-		free(out);
-	}
-	return 0;
-
-_err_:
-	free(out);
-#ifndef NO_CPACF
-	syslog(LOG_ERR, "Libica 3DES-CBC-CS test failed.");
-#endif
-	return 1;
-}
-
-static int
-des3_cfb_kat(void) {
-	const struct des3_cfb_tv *tv;
-	size_t i;
-	unsigned char iv[DES3_BLKSIZE], *out;
-
-	for (i = 0; i < DES3_CFB_TV_LEN; i++) {
-		tv = &DES3_CFB_TV[i];
-
-		out = malloc(tv->len);
-
-		if (out == NULL)
-			goto _err_;
-
-		memcpy(iv, tv->iv, DES3_BLKSIZE);
-		if (ica_3des_cfb(tv->plaintext, out, tv->len,
-		    (unsigned char *)tv->key, iv, tv->lcfb, ICA_ENCRYPT)
-		    || memcmp(tv->ciphertext, out, tv->len))
-			goto _err_;
-
-		memcpy(iv, tv->iv, DES3_BLKSIZE);
-		if (ica_3des_cfb(tv->ciphertext, out, tv->len,
-		    (unsigned char *)tv->key, iv, tv->lcfb, ICA_DECRYPT)
-		    || memcmp(tv->plaintext, out, tv->len))
-			goto _err_;
-
-		free(out);
-	}
-	return 0;
-
-_err_:
-	free(out);
-#ifndef NO_CPACF
-	syslog(LOG_ERR, "Libica 3DES-CFB test failed.");
-#endif
-	return 1;
-}
-
-static int
-des3_ofb_kat(void) {
-	const struct des3_ofb_tv *tv;
-	size_t i;
-	unsigned char iv[DES3_BLKSIZE], *out;
-
-	for (i = 0; i < DES3_OFB_TV_LEN; i++) {
-		tv = &DES3_OFB_TV[i];
-
-		out = malloc(tv->len);
-
-		if (out == NULL)
-			goto _err_;
-
-		memcpy(iv, tv->iv, DES3_BLKSIZE);
-		if (ica_3des_ofb(tv->plaintext, out, tv->len,
-		    (unsigned char *)tv->key, iv, ICA_ENCRYPT)
-		    || memcmp(tv->ciphertext, out, tv->len))
-			goto _err_;
-
-		memcpy(iv, tv->iv, DES3_BLKSIZE);
-		if (ica_3des_ofb(tv->ciphertext, out, tv->len,
-		    (unsigned char *)tv->key, iv, ICA_DECRYPT)
-		    || memcmp(tv->plaintext, out, tv->len))
-			goto _err_;
-
-		free(out);
-	}
-	return 0;
-
-_err_:
-	free(out);
-#ifndef NO_CPACF
-	syslog(LOG_ERR, "Libica 3DES-OFB test failed.");
-#endif
-	return 1;
-}
-
-static int
-des3_ctr_kat(void) {
-	const struct des3_ctr_tv *tv;
-	size_t i;
-	unsigned char *out, ctr[DES3_BLKSIZE];
-
-	for (i = 0; i < DES3_CTR_TV_LEN; i++) {
-		tv = &DES3_CTR_TV[i];
-
-		out = malloc(tv->len);
-
-		if (out == NULL)
-			goto _err_;
-
-		memcpy(ctr, tv->ctr, DES3_BLKSIZE);
-		if (ica_3des_ctr(tv->plaintext, out, tv->len,
-		    (unsigned char *)tv->key, ctr, 32, ICA_ENCRYPT)
-		    || memcmp(tv->ciphertext, out, tv->len))
-			goto _err_;
-
-		memcpy(ctr, tv->ctr, DES3_BLKSIZE);
-		if (ica_3des_ctr(tv->ciphertext, out, tv->len,
-		    (unsigned char *)tv->key, ctr, 32, ICA_DECRYPT)
-		    || memcmp(tv->plaintext, out, tv->len))
-			goto _err_;
-
-		free(out);
-	}
-	return 0;
-
-_err_:
-	free(out);
-#ifndef NO_CPACF
-	syslog(LOG_ERR, "Libica 3DES-CTR test failed.");
-#endif
-	return 1;
-}
-
-static int
-des3_cmac_kat(void)
+ecdh_kat()
 {
-	const struct des3_cmac_tv *tv;
-	size_t i, lastlen;
-	unsigned char *mac, iv[DES3_BLKSIZE];
+	ica_adapter_handle_t ah;
+	const struct ecdh_kat_tv *tv;
+	unsigned char shared_secret[MAX_ECC_PRIV_SIZE];
+	ICA_EC_KEY *eckey_A, *eckey_B;
+	unsigned int privlen;
+	size_t i;
 
-	for (i = 0; i < DES3_CMAC_TV_LEN; i++) {
-		tv = &DES3_CMAC_TV[i];
+	if (!function_supported_via_cpacf(EC_DH))
+		return 0;
 
-		mac = malloc(tv->maclen);
+	if (ica_open_adapter(&ah))
+		return 1;
 
-		if (mac == NULL)
+	for (i = 0; i < ECDH_KAT_TV_LEN; i++) {
+
+		tv = &ECDH_KAT_TV[i];
+
+		eckey_A = ica_ec_key_new(tv->nid, &privlen);
+		eckey_B = ica_ec_key_new(tv->nid, &privlen);
+		if (!eckey_A || !eckey_B)
 			goto _err_;
 
-		if ((ica_3des_cmac(tv->msg, tv->msglen, tv->mac, tv->maclen,
-		    (unsigned char *)tv->key, ICA_DECRYPT) != tv->rv)
-		    || ((tv->rv == 0) && (ica_3des_cmac(tv->msg, tv->msglen,
-		    mac, tv->maclen, (unsigned char *)tv->key, ICA_ENCRYPT)
-		    || memcmp(tv->mac, mac, tv->maclen))))
+		/* calculate shared secret with pub_B, priv_A */
+		if (ica_ec_key_init(NULL, NULL, tv->da, eckey_A))
 			goto _err_;
 
-		free(mac);
+		if (ica_ec_key_init(tv->xb, tv->yb, NULL, eckey_B))
+			goto _err_;
+
+		memset(shared_secret, 0, sizeof(shared_secret));
+		if (ica_ecdh_derive_secret(ah, eckey_A, eckey_B, shared_secret, privlen))
+			goto _err_;
+
+		/* compare result with known result */
+		if (memcmp(shared_secret, tv->z, tv->privlen) != 0)
+			goto _err_;
+
+		ica_ec_key_free(eckey_A);
+		ica_ec_key_free(eckey_B);
+
+		eckey_A = ica_ec_key_new(tv->nid, &privlen);
+		eckey_B = ica_ec_key_new(tv->nid, &privlen);
+		if (!eckey_A || !eckey_B)
+			goto _err_;
+
+		/* calculate shared secret with pub_A, priv_B */
+		if (ica_ec_key_init(NULL, NULL, tv->db, eckey_B))
+			goto _err_;
+
+		if (ica_ec_key_init(tv->xa, tv->ya, NULL, eckey_A))
+			goto _err_;
+
+		memset(shared_secret, 0, sizeof(shared_secret));
+		if (ica_ecdh_derive_secret(ah, eckey_B, eckey_A, shared_secret, privlen))
+			goto _err_;
+
+		/* compare result with known result */
+		if (memcmp(shared_secret, tv->z, tv->privlen) != 0)
+			goto _err_;
+
+		ica_ec_key_free(eckey_A);
+		ica_ec_key_free(eckey_B);
 	}
-	for (i = 0; i < DES3_CMAC_TV_LEN; i++) {
-		tv = &DES3_CMAC_TV[i];
 
-		if (tv->msglen <= DES3_BLKSIZE)
-			continue;
-
-		lastlen = tv->msglen % DES3_BLKSIZE ? tv->msglen % DES3_BLKSIZE
-		    : DES3_BLKSIZE;
-		memset(iv, 0, DES3_BLKSIZE);
-
-		if (ica_3des_cmac_intermediate(tv->msg, tv->msglen - lastlen,
-		    (unsigned char *)tv->key, iv)
-		    || (ica_3des_cmac_last(tv->msg + (tv->msglen - lastlen),
-		    lastlen, tv->mac, tv->maclen, (unsigned char *)tv->key, iv,
-		    ICA_DECRYPT) != tv->rv))
-			return 1;
-
-		if (tv->rv != 0)
-			continue;
-
-		mac = malloc(tv->maclen);
-
-		if (mac == NULL)
-			goto _err_;
-
-		memset(iv, 0, DES3_BLKSIZE);
-		if (ica_3des_cmac_intermediate(tv->msg, tv->msglen - lastlen,
-		    (unsigned char *)tv->key, iv)
-		    || ica_3des_cmac_last(tv->msg + (tv->msglen - lastlen),
-		    lastlen, mac, tv->maclen, (unsigned char *)tv->key, iv,
-		    ICA_ENCRYPT) || memcmp(tv->mac, mac, tv->maclen))
-			goto _err_;
-
-		free(mac);
-	}
+	ica_close_adapter(ah);
 	return 0;
 
 _err_:
-	free(mac);
-#ifndef NO_CPACF
-	syslog(LOG_ERR, "Libica 3DES-CMAC test failed.");
-#endif
+	ica_ec_key_free(eckey_A);
+	ica_ec_key_free(eckey_B);
+	ica_close_adapter(ah);
+	syslog(LOG_ERR, "Libica ECDH test failed.");
+	return 1;
+}
+
+static int
+ecdsa_kat(void)
+{
+	ICA_EC_KEY *eckey;
+	const struct ecdsa_kat_tv *tv;
+	unsigned char sigbuf[MAX_ECDSA_SIG_SIZE];
+	unsigned int privlen;
+	size_t i;
+	int rc;
+
+	if (!function_supported_via_cpacf(EC_DSA_SIGN))
+		return 0;
+
+	for (i = 0; i < ECDSA_KAT_TV_LEN; i++) {
+		tv = &ECDSA_KAT_TV[i];
+		eckey = ica_ec_key_new(tv->nid, &privlen);
+		if (!eckey)
+			goto _err_;
+		rc = ica_ec_key_init(tv->x, tv->y, tv->d, eckey);
+		if (rc)
+			goto _err_;
+		/* adapter handle not needed here, just CPACF */
+		rc = ica_ecdsa_sign_ex_internal(0, eckey, tv->hash, tv->hashlen,
+								sigbuf, tv->siglen, tv->k);
+		if (rc)
+			goto _err_;
+		rc = ica_ecdsa_verify(0, eckey, tv->hash, tv->hashlen, sigbuf, tv->siglen);
+		if (rc)
+			goto _err_;
+		if (memcmp(sigbuf, tv->sig, tv->siglen) != 0) {
+			goto _err_;
+		}
+		ica_ec_key_free(eckey);
+	}
+
+	return 0;
+
+_err_:
+	ica_ec_key_free(eckey);
+	syslog(LOG_ERR, "Libica ECDSA test failed.");
 	return 1;
 }
 
