@@ -66,9 +66,13 @@ static unsigned int mod_expo_sw(int arg_length, unsigned char *arg, int exp_leng
 RSA* rsa_key_generate(unsigned int modulus_bit_length,
 		      unsigned long *public_exponent)
 {
+	int min_pubexp = 3;
+
 #ifdef ICA_FIPS
 	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return NULL;
+	if (fips & ICA_FIPS_MODE)
+		min_pubexp = 65537;
 #endif /* ICA_FIPS */
 
 	if (*public_exponent == 0)
@@ -77,7 +81,7 @@ RSA* rsa_key_generate(unsigned int modulus_bit_length,
 			if (s390_prng((unsigned char*)public_exponent,
 			    sizeof(unsigned long)) != 0)
 				return NULL;
-		} while (*public_exponent <= 2 || !(*public_exponent % 2));
+		} while (*public_exponent < min_pubexp || !(*public_exponent % 2));
 	}
 
 	RSA *rsa = RSA_new();
@@ -113,6 +117,7 @@ EVP_PKEY* rsa_key_generate(unsigned int modulus_bit_length,
 	EVP_PKEY *pkey = NULL;
 	EVP_PKEY_CTX *pctx = NULL;
 	BIGNUM *e = NULL;
+	int try;
 
 #ifdef ICA_FIPS
 	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
@@ -145,7 +150,14 @@ EVP_PKEY* rsa_key_generate(unsigned int modulus_bit_length,
 		goto done;
 	}
 
-	if (EVP_PKEY_keygen(pctx, &pkey) != 1) {
+	/*
+	 * In OpenSSL 3.0 the RSA key gen algorithm has been changed and can now
+	 * fail to generate a key. Retry up to 10 times in such a case.
+	 */
+	for (try = 1; try <= 10; try++) {
+		if (EVP_PKEY_keygen(pctx, &pkey) == 1)
+			break;
+
 		if (pkey)
 			EVP_PKEY_free(pkey);
 		pkey = NULL;
@@ -189,10 +201,8 @@ unsigned int rsa_key_generate_mod_expo(ica_adapter_handle_t deviceHandle,
 #ifdef ICA_FIPS
 	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
-#if OPENSSL_VERSION_PREREQ(3, 0)
-	if ((fips & ICA_FIPS_MODE) && (modulus_bit_length <= 1024))
+	if ((fips & ICA_FIPS_MODE) && (modulus_bit_length < 2048))
 		return EPERM;
-#endif
 #endif /* ICA_FIPS */
 
 	BEGIN_OPENSSL_LIBCTX(openssl_libctx, rc);
@@ -294,10 +304,8 @@ unsigned int rsa_key_generate_crt(ica_adapter_handle_t deviceHandle,
 #ifdef ICA_FIPS
 	if ((fips & ICA_FIPS_MODE) && (!openssl_in_fips_mode()))
 		return EACCES;
-#if OPENSSL_VERSION_PREREQ(3, 0)
-	if ((fips & ICA_FIPS_MODE) && (modulus_bit_length <= 1024))
+	if ((fips & ICA_FIPS_MODE) && (modulus_bit_length < 2048))
 		return EPERM;
-#endif
 #endif /* ICA_FIPS */
 
 	BEGIN_OPENSSL_LIBCTX(openssl_libctx, rc);
